/*
  tcp_seg.c

  Dug Song <dugsong@anzen.com>

  Copyright (c) 1999 Anzen Computing. All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
  
  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  3. All advertising materials mentioning features or use of this software
     must display the following acknowledgement:
     This product includes software developed by Anzen Computing.
  4. Neither the name of Anzen Computing nor the names of its
     contributors may be used to endorse or promote products derived
     from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
  IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
  ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  $Id: tcp_seg.c,v 1.1 1999/05/18 15:34:06 dugsong Exp $
*/

#include <libnet.h>
#include "list.h"
#include "send.h"
#include "tcp_seg.h"

ELEM *
tcp_seg_make(u_char *pkt, int pktlen, int segsize)
{
  ELEM *new, *list = NULL;
  struct ip *iph = (struct ip *)pkt;
  int len, ip_hl = iph->ip_hl * 4;
  struct tcp *tcph = (struct tcp *)(pkt + (iph->ip_hl * 4));
  int tcp_hl = tcph->th_off * 4;
  u_char *tcp_blob = pkt + ip_hl + tcp_hl;
  int tcp_bloblen = ntohs(iph->ip_len) - ip_hl - tcp_hl;
  u_char *tcp_end = tcp_blob + tcp_bloblen;
  u_long tcp_newseq = ntohl(tcph->th_seq);
  u_char *p, *data;

  if (iph->ip_p != IPPROTO_TCP || tcp_bloblen <= segsize ||
      (tcph->th_flags & TH_ACK) == 0)
    return NULL;

  for (p = tcp_blob ; p < tcp_end ; p += len) {
    len = MIN(tcp_end - p, segsize);
    
    if (libnet_init_packet(ip_hl + tcp_hl + len, &data) == -1)
      return NULL;

    /* Copy in headers and data. */
    memcpy(data, pkt, ip_hl + tcp_hl);
    memcpy(data + ip_hl + tcp_hl, p, len);

    /* Correct IP len, TCP seqnum, and TCP checksum. */
    ((struct ip *)data)->ip_len = htons(ip_hl + tcp_hl + len);
    ((struct tcp *)(data + ip_hl))->th_seq = htonl(tcp_newseq);
    libnet_do_checksum(data, IPPROTO_TCP, tcp_hl + len);
    
    tcp_newseq += len;

    /* Add it to our list of TCP packets. */
    new = list_elem(data, ip_hl + tcp_hl + len);
    
    if (!(list = list_add(list, new)))
      return NULL;
  }
  return (list->head);
}
    
ELEM *
tcp_seg_fake_close(u_char *pkt, int pktlen)
{
  ELEM *new, *list = NULL;
  struct ip *iph = (struct ip *)pkt;
  int ip_hl = iph->ip_hl * 4;
  struct tcp *tcph = (struct tcp *)(pkt + ip_hl);
  u_char *data;

  if ((tcph->th_flags & TH_SYN) == 0)
    return NULL;

  /* Add our real SYN packet to list. */
  new = list_elem(pkt, pktlen);
  if (!(list = list_add(list, new)))
    return NULL;
  
  /* Add our fake FIN packet. */
  if (libnet_init_packet(ip_hl + TCP_H, &data) == -1)
    return NULL;

  memcpy(data, pkt, ip_hl);
  ((struct ip *)data)->ip_len = htons(ip_hl + TCP_H);
  ((struct ip *)data)->ip_id = htons(iph->ip_id); /* intentional! */
  libnet_build_tcp(ntohs(tcph->th_sport), ntohs(tcph->th_dport),
		   ntohl(tcph->th_seq) + 1, ntohl(tcph->th_ack),
		   TH_FIN, ntohs(tcph->th_win), ntohs(tcph->th_urp),
		   NULL, 0, data + ip_hl);

  new = list_elem(data, ip_hl + TCP_H);
  if (!(list = list_add(list, new)))
          return NULL;

  /* Add our fake RST packet. */
  if (libnet_init_packet(ip_hl + TCP_H, &data) == -1)
    return NULL;

  memcpy(data, pkt, ip_hl);
  ((struct ip *)data)->ip_len = htons(ip_hl + TCP_H);
  ((struct ip *)data)->ip_id = htons(iph->ip_id); /* intentional! */
  libnet_build_tcp(ntohs(tcph->th_sport), ntohs(tcph->th_dport),
		   ntohl(tcph->th_seq) + 2, ntohl(tcph->th_ack),
		   TH_RST, ntohs(tcph->th_win), ntohs(tcph->th_urp),
		   NULL, 0, data + ip_hl);

  new = list_elem(data, ip_hl + TCP_H);
  if (!(list = list_add(list, new)))
    return NULL;
  
  return (list->head);
}

ELEM *
tcp_seg_null_payload(ELEM *seg)
{
  struct ip *iph = (struct ip *)seg->data;
  int ip_hl = iph->ip_hl * 4;
  struct tcp *tcph = (struct tcp *)(seg->data + ip_hl);
  int tcp_hl = tcph->th_off * 4;
  int newlen = ntohs(iph->ip_len) - (ip_hl + tcp_hl);
  u_char *newdata;
  
  if (libnet_init_packet(ip_hl + TCP_H + newlen, &newdata) == -1)
    return NULL;
  
  /* Whack the payload (including TCP options to elude PAWS). */
  memcpy(newdata, seg->data, ip_hl + TCP_H);
  memset(newdata + ip_hl + TCP_H, 0, newlen);

  /* Correct IP length, TCP header length, TCP checksum, segment length. */
  ((struct ip *)newdata)->ip_len = htons(ip_hl + TCP_H + newlen);
  ((struct tcp *)(newdata + ip_hl))->th_off = TCP_H / 4;
  libnet_do_checksum(newdata, IPPROTO_TCP, TCP_H + newlen);
  seg->len = ip_hl + TCP_H + newlen;

  free(seg->data);
  seg->data = newdata;
  
  return (seg);
}

ELEM *
tcp_seg_whack_checksums(ELEM *seg)
{
  ELEM *s;
  struct ip *iph;
  struct tcp *tcph;

  for (s = seg ; s != NULL ; s = s->next) {
    iph = (struct ip *)s->data;
    tcph = (struct tcp *)(s->data + (iph->ip_hl * 4));
    tcph->th_sum = 666;
  }  
  return (seg);
}

ELEM *
tcp_seg_whack_acks(ELEM *seg)
{
  ELEM *s;
  struct ip *iph;
  struct tcp *tcph;

  for (s = seg ; s != NULL ; s = s->next) {
    iph = (struct ip *)s->data;
    tcph = (struct tcp *)(s->data + (iph->ip_hl * 4));
    tcph->th_flags &= ~TH_ACK;
  }  
  return (seg);
}

ELEM *
tcp_seg_interleave_nulls(ELEM *seg)
{
  ELEM *s;
  struct ip *iph;
  int ip_hl;

  if (!seg) return NULL;

  for (s = seg ; s != NULL ; s = s->next) {
    s = list_dup(s);

    iph = (struct ip *)s->data;
    ip_hl = iph->ip_hl * 4;
    
    /* Whack the sequence number with something bogus. */
    ((struct tcp *)(s->data + ip_hl))->th_seq =
      htonl(((struct tcp *)(s->data + ip_hl))->th_seq);

    /* Whack the payload. */
    tcp_seg_null_payload(s);
  }
  return (seg);
}

ELEM *
tcp_seg_interleave_syns(ELEM *seg)
{
  ELEM *s, *new;
  u_char *data;

  if (!seg) return NULL;

  for (s = seg ; s && s->next ; s = s->next) {
    struct ip *iph = (struct ip *)s->data;
    int ip_hl = iph->ip_hl * 4;
    struct tcp *tcph = (struct tcp *)(s->data + ip_hl);

    if (libnet_init_packet(ip_hl + TCP_H, &data) == -1)
      return NULL;

    /* Copy in IP header. */
    memcpy(data, s->data, ip_hl);

    /* Correct IP length, IP ID. */
    ((struct ip *)data)->ip_len = htons(ip_hl + TCP_H);
    ((struct ip *)data)->ip_id = htons(iph->ip_id); /* intentional! */

    libnet_build_tcp(ntohs(tcph->th_sport), ntohs(tcph->th_dport),
		     htonl(tcph->th_seq), 0, TH_SYN, ntohs(tcph->th_win),
		     ntohs(tcph->th_urp), NULL, 0, data + ip_hl);
    
    libnet_do_checksum(data, IPPROTO_TCP, TCP_H);

    /* Add it to our list of TCP packets. */
    new = list_elem(data, ip_hl + TCP_H);
    
    if (!(s = list_add(s, new)))
      return NULL;
  }
  return (seg);
}

ELEM *
tcp_seg_prepend_junk(u_char *pkt, int pktlen)
{
  ELEM *new, *list = NULL;
  struct ip *iph = (struct ip *)pkt;
  int ip_hl = iph->ip_hl * 4;
  struct tcp *tcph = (struct tcp *)(pkt + ip_hl);
  u_char *data, nulls[PACKET];
  int i;

  if ((tcph->th_flags & TH_SYN) == 0)
    return NULL;

  memset(nulls, 0, PACKET);
  
  for (i = 0; i < 500 ; i++) {
    if (libnet_init_packet(ip_hl + TCP_H + PACKET, &data) == -1)
      return NULL;

    memcpy(data, pkt, ip_hl);
    ((struct ip *)data)->ip_len = htons(ip_hl + TCP_H + PACKET);
    ((struct ip *)data)->ip_id = htons(iph->ip_id + i); /* intentional! */
    libnet_build_tcp(ntohs(tcph->th_sport), ntohs(tcph->th_dport),
		     i, 666, TH_ACK, ntohs(tcph->th_win), ntohs(tcph->th_urp),
		     nulls, PACKET, data + ip_hl);
    /*    libnet_do_checksum(data, IPPROTO_TCP, TCP_H + PACKET);*/
    
    new = list_elem(data, ip_hl + TCP_H + PACKET);
    if (!(list = list_add(list, new)))
      return NULL;
  }
  /* Add our real SYN packet. */
  new = list_elem(pkt, pktlen);
  if (!(list = list_add(list, new)))
    return NULL;
  
  return (list->head);
}

ELEM *
tcp_seg_interleave_overwrites(ELEM *seg)
{
  ELEM *s, *new;
  u_char nulls[PACKET], *newdata;
  u_long newseq;
  int newlen;

  memset(nulls, 0, sizeof(nulls));
  
  for (s = seg ; s && s->next ; s = s->next) {
    struct ip *iph = (struct ip *)s->data;
    int ip_hl = iph->ip_hl * 4;
    struct tcp *tcph = (struct tcp *)(s->data + ip_hl);
    int tcp_hl = tcph->th_off * 4;

    /* Build new null data segment, half the length of the current segment
       and overlapping its latter half. */
    newlen = (ntohs(iph->ip_len) - (ip_hl + tcp_hl)) / 2;
    newseq = ntohl(tcph->th_seq) - newlen;

    if (libnet_init_packet(ip_hl + TCP_H + newlen, &newdata) == -1)
      return NULL;

    /* Copy in IP, TCP headers and data. */
    memcpy(newdata, s->data, ip_hl + TCP_H);
    memcpy(newdata + ip_hl + TCP_H, nulls, newlen);

    /* Correct IP length, TCP header length, TCP checksum. */
    ((struct ip *)newdata)->ip_len = htons(ip_hl + TCP_H + newlen);
    ((struct tcp *)(newdata + ip_hl))->th_off = TCP_H / 4; /* XXX - PAWS! */
    libnet_do_checksum(newdata, IPPROTO_TCP, TCP_H + newlen);

    /* Add null fragment after current fragment. */
    new = list_elem(newdata, ip_hl + TCP_H + newlen);
    
    if (!list_add(s, new))
      return NULL;
    
    /* Swap them. */
    if (!list_swap(s))
      return NULL;
  }
  return (seg);
}
